/*  gtkform.c
 *  Copyright (C) 2001 Steve Green
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * 7 Jul 2005 - Ron Yorston <rmy@tigress.co.uk>
 *		o  Updated from java version for "lower_container_relative".
 * 		o  Added ATTACH_SELF.
 *		o  Updated for GTK+ 2.x.
 *		o  Test program
 */

/*
 * TODO
 *
 * Remove stdio (printf)
 * More convience functions
 */

#include "gtkform.h"
#include <setjmp.h>
#include <stdio.h>

enum
{
  STATE_RESET,
  STATE_VISITED,
  STATE_DONE
};

enum
{
  CHILD_PROP_TOP_ATTACHMENT = 1,
  CHILD_PROP_TOP_WIDGET,
  CHILD_PROP_TOP_OFFSET,
  CHILD_PROP_LEFT_ATTACHMENT,
  CHILD_PROP_LEFT_WIDGET,
  CHILD_PROP_LEFT_OFFSET,
  CHILD_PROP_BOTTOM_ATTACHMENT,
  CHILD_PROP_BOTTOM_WIDGET,
  CHILD_PROP_BOTTOM_OFFSET,
  CHILD_PROP_RIGHT_ATTACHMENT,
  CHILD_PROP_RIGHT_WIDGET,
  CHILD_PROP_RIGHT_OFFSET,
};

static void gtk_form_class_init    (GtkFormClass  *klass);
static void gtk_form_init	    (GtkForm	    *form);
static void gtk_form_finalize	    (GObject	    *object);
static void gtk_form_size_request  (GtkWidget	    *widget,
				     GtkRequisition *requisition);
static void gtk_form_size_allocate (GtkWidget	    *widget,
				     GtkAllocation  *allocation);
static void gtk_form_add	    (GtkContainer   *container,
				     GtkWidget	    *widget);
static void gtk_form_remove	    (GtkContainer   *container,
				     GtkWidget	    *widget);
static void gtk_form_forall	    (GtkContainer   *container,
				     gboolean	     include_internals,
				     GtkCallback     callback,
				     gpointer	     callback_data);
static void gtk_form_set_child_property (GtkContainer   *container,
				     GtkWidget      *child,
				     guint           property_id,
				     const GValue   *value,
				     GParamSpec     *pspec);
static void gtk_form_get_child_property (GtkContainer   *container,
				     GtkWidget      *child,
				     guint           property_id,
				     GValue         *value,
				     GParamSpec     *pspec);
static GtkType gtk_form_child_type (GtkContainer   *container);


static void gtk_form_layout	   (GtkForm	   *form,
				    gint	   *my_bounds,
				    gboolean	    recompute_our_size);
static void gtk_form_layout_child  (jmp_buf	    env,
				    GtkFormChild   *fc,
				    gint           *my_bounds,
				    gboolean	    recompute_our_size);
static gint gtk_form_layout_edge   (jmp_buf         env,
				    GtkFormChild   *fc,
				    gint            edge,
				    gint           *my_bounds,
				    gboolean        recompute_our_size);
static void gtk_form_move_edge     (jmp_buf         env,
                                    GtkFormChild   *fc,
				    int             edge,
				    int             where,
				    gint           *my_bounds,
				    gboolean        recompute);
static gboolean gtk_form_edge_should_move_too (GtkFormChild *fc,
						gint edge);

static GtkContainerClass *parent_class = NULL;


static GType
gtk_form_attachment_get_type(void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { GTK_FORM_ATTACH_NONE, "GTK_FORM_ATTACH_NONE", "none" },
      { GTK_FORM_ATTACH_FORM, "GTK_FORM_ATTACH_FORM", "form" },
      { GTK_FORM_ATTACH_WIDGET, "GTK_FORM_ATTACH_WIDGET", "widget" },
      { GTK_FORM_ATTACH_OPPOSITE_WIDGET, "GTK_FORM_ATTACH_OPPOSITE_WIDGET", "opposite_widget" },
      { GTK_FORM_ATTACH_CENTER, "GTK_FORM_ATTACH_CENTER", "center" },
	  { GTK_FORM_ATTACH_SELF, "GTK_FORM_ATTACH_SELF", "self" },
      { 0, NULL, NULL }
    };
	etype = g_enum_register_static ("GtkFormAttachment", values);
  }
  return etype;
}
#define GTK_TYPE_FORM_ATTACHMENT (gtk_form_attachment_get_type())

GType
gtk_form_get_type (void)
{
  static GType form_type = 0;
  
  if (!form_type)
    {
      static const GTypeInfo form_info =
      {
	sizeof (GtkFormClass),
	NULL,
	NULL,
	(GClassInitFunc) gtk_form_class_init,
	NULL,
	NULL,
	sizeof (GtkForm),
	0,
	(GInstanceInitFunc) gtk_form_init
      };
      
      form_type = g_type_register_static (GTK_TYPE_CONTAINER, "GtkForm",
	  				&form_info, 0);
    }
  
  return form_type;
}

static void
gtk_form_class_init (GtkFormClass *class)
{
  GObjectClass *object_class = G_OBJECT_CLASS (class);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
  GtkContainerClass *container_class = GTK_CONTAINER_CLASS (class);
  
  parent_class = g_type_class_peek_parent (class);
  
  object_class->finalize = gtk_form_finalize;

  widget_class->size_request = gtk_form_size_request;
  widget_class->size_allocate = gtk_form_size_allocate;
  
  container_class->add = gtk_form_add;
  container_class->remove = gtk_form_remove;
  container_class->forall = gtk_form_forall;
  container_class->child_type = gtk_form_child_type;
  container_class->set_child_property = gtk_form_set_child_property;
  container_class->get_child_property = gtk_form_get_child_property;

  gtk_container_class_install_child_property (container_class,
  						CHILD_PROP_TOP_ATTACHMENT,
						g_param_spec_enum ("top_attachment",
							"Top attachment",
							"Type of attachment of top edge of child",
							GTK_TYPE_FORM_ATTACHMENT, GTK_FORM_ATTACH_NONE,
							G_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
  						CHILD_PROP_TOP_WIDGET,
						g_param_spec_object ("top_widget",
							"Top widget",
							"Target widget for top edge of child",
							GTK_TYPE_WIDGET,
							G_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
  						CHILD_PROP_TOP_OFFSET,
						g_param_spec_uint ("top_offset",
							"Top offset",
							"Offset from the target or form side",
							0, G_MAXUINT, 0,
							G_PARAM_READWRITE));

  gtk_container_class_install_child_property (container_class,
  						CHILD_PROP_LEFT_ATTACHMENT,
						g_param_spec_enum ("left_attachment",
							"Left attachment",
							"Type of attachment of left edge of child",
							GTK_TYPE_FORM_ATTACHMENT, GTK_FORM_ATTACH_NONE,
							G_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
  						CHILD_PROP_LEFT_WIDGET,
						g_param_spec_object ("left_widget",
							"Left widget",
							"Target widget for left edge of child",
							GTK_TYPE_WIDGET,
							G_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
  						CHILD_PROP_LEFT_OFFSET,
						g_param_spec_uint ("left_offset",
							"Left offset",
							"Offset from the target or form side",
							0, G_MAXUINT, 0,
							G_PARAM_READWRITE));

  gtk_container_class_install_child_property (container_class,
  						CHILD_PROP_BOTTOM_ATTACHMENT,
						g_param_spec_enum ("bottom_attachment",
							"Bottom attachment",
							"Type of attachment of bottom edge of child",
							GTK_TYPE_FORM_ATTACHMENT, GTK_FORM_ATTACH_NONE,
							G_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
  						CHILD_PROP_BOTTOM_WIDGET,
						g_param_spec_object ("bottom_widget",
							"Bottom widget",
							"Target widget for bottom edge of child",
							GTK_TYPE_WIDGET,
							G_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
  						CHILD_PROP_BOTTOM_OFFSET,
						g_param_spec_uint ("bottom_offset",
							"Bottom offset",
							"Offset from the target or form side",
							0, G_MAXUINT, 0,
							G_PARAM_READWRITE));

  gtk_container_class_install_child_property (container_class,
  						CHILD_PROP_RIGHT_ATTACHMENT,
						g_param_spec_enum ("right_attachment",
							"Right attachment",
							"Type of attachment of right edge of child",
							GTK_TYPE_FORM_ATTACHMENT, GTK_FORM_ATTACH_NONE,
							G_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
  						CHILD_PROP_RIGHT_WIDGET,
						g_param_spec_object ("right_widget",
							"Right widget",
							"Target widget for right edge of child",
							GTK_TYPE_WIDGET,
							G_PARAM_READWRITE));
  gtk_container_class_install_child_property (container_class,
  						CHILD_PROP_RIGHT_OFFSET,
						g_param_spec_uint ("right_offset",
							"Right offset",
							"Offset from the target or form side",
							0, G_MAXUINT, 0,
							G_PARAM_READWRITE));
}

static GtkType
gtk_form_child_type (GtkContainer   *container)
{
  return GTK_TYPE_WIDGET;
}

static GtkFormChild *
gtk_form_find_child (GtkForm *form, GtkWidget *widget)
{
  GList *list;
  for (list = form->children; list; list = list->next)
  {
      GtkFormChild *form_child = list->data;
      if (form_child->widget == widget)
	return form_child;
  }
  return NULL;
}

static void
gtk_form_set_child_property (GtkContainer      *container,
			GtkWidget    *child,
			guint         property_id,
			const GValue *value,
			GParamSpec   *pspec)
{
  GtkForm *form = GTK_FORM(container);
  GtkFormChild *form_child;

  form_child = gtk_form_find_child (form, child);
  if (!form_child) {
	GTK_CONTAINER_WARN_INVALID_CHILD_PROPERTY_ID (container, property_id, pspec);
    return;
  }

  switch (property_id) {
  case CHILD_PROP_TOP_ATTACHMENT:
    form_child->constraints [GTK_FORM_EDGE_TOP].attachment = g_value_get_enum(value);
	break;
  case CHILD_PROP_TOP_WIDGET:
    form_child->constraints [GTK_FORM_EDGE_TOP].child = gtk_form_find_child (form, GTK_WIDGET (g_value_get_object (value)));
	break;
  case CHILD_PROP_TOP_OFFSET:
    form_child->constraints [GTK_FORM_EDGE_TOP].offset = g_value_get_uint(value);
	break;
  case CHILD_PROP_LEFT_ATTACHMENT:
    form_child->constraints [GTK_FORM_EDGE_LEFT].attachment = g_value_get_enum(value);
	break;
  case CHILD_PROP_LEFT_WIDGET:
    form_child->constraints [GTK_FORM_EDGE_LEFT].child = gtk_form_find_child (form, GTK_WIDGET (g_value_get_object (value)));
	break;
  case CHILD_PROP_LEFT_OFFSET:
    form_child->constraints [GTK_FORM_EDGE_LEFT].offset = g_value_get_uint(value);
	break;
  case CHILD_PROP_BOTTOM_ATTACHMENT:
    form_child->constraints [GTK_FORM_EDGE_BOTTOM].attachment = g_value_get_enum(value);
	break;
  case CHILD_PROP_BOTTOM_WIDGET:
    form_child->constraints [GTK_FORM_EDGE_BOTTOM].child = gtk_form_find_child (form, GTK_WIDGET (g_value_get_object (value)));
	break;
  case CHILD_PROP_BOTTOM_OFFSET:
    form_child->constraints [GTK_FORM_EDGE_BOTTOM].offset = g_value_get_uint(value);
	break;
  case CHILD_PROP_RIGHT_ATTACHMENT:
    form_child->constraints [GTK_FORM_EDGE_RIGHT].attachment = g_value_get_enum(value);
	break;
  case CHILD_PROP_RIGHT_WIDGET:
    form_child->constraints [GTK_FORM_EDGE_RIGHT].child = gtk_form_find_child (form, GTK_WIDGET (g_value_get_object (value)));
	break;
  case CHILD_PROP_RIGHT_OFFSET:
    form_child->constraints [GTK_FORM_EDGE_RIGHT].offset = g_value_get_uint(value);
	break;
  default:
    GTK_CONTAINER_WARN_INVALID_CHILD_PROPERTY_ID (container, property_id, pspec);
    break;
  }
}

static void
gtk_form_get_child_property (GtkContainer      *container,
			GtkWidget    *child,
			guint         property_id,
			GValue       *value,
			GParamSpec   *pspec)
{
  GtkForm *form = GTK_FORM(container);
  GtkFormChild *form_child;

  form_child = gtk_form_find_child (form, child);
  if (!form_child) {
	GTK_CONTAINER_WARN_INVALID_CHILD_PROPERTY_ID (container, property_id, pspec);
    return;
  }

  switch (property_id) {
  case CHILD_PROP_TOP_ATTACHMENT:
    g_value_set_enum (value, form_child->constraints[GTK_FORM_EDGE_TOP].attachment);
	break;
  case CHILD_PROP_TOP_WIDGET:
    g_value_set_object (value,
		form_child->constraints[GTK_FORM_EDGE_TOP].child ?
		G_OBJECT(form_child->constraints[GTK_FORM_EDGE_TOP].child) : NULL);
    break;
  case CHILD_PROP_TOP_OFFSET:
    g_value_set_uint(value, form_child->constraints[GTK_FORM_EDGE_TOP].offset);
	break;
  case CHILD_PROP_LEFT_ATTACHMENT:
    g_value_set_enum (value, form_child->constraints[GTK_FORM_EDGE_LEFT].attachment);
	break;
  case CHILD_PROP_LEFT_WIDGET:
    g_value_set_object (value,
		form_child->constraints[GTK_FORM_EDGE_LEFT].child ?
		G_OBJECT(form_child->constraints[GTK_FORM_EDGE_LEFT].child) : NULL);
    break;
  case CHILD_PROP_LEFT_OFFSET:
    g_value_set_uint(value, form_child->constraints[GTK_FORM_EDGE_LEFT].offset);
	break;
  case CHILD_PROP_BOTTOM_ATTACHMENT:
    g_value_set_enum (value, form_child->constraints[GTK_FORM_EDGE_BOTTOM].attachment);
	break;
  case CHILD_PROP_BOTTOM_WIDGET:
    g_value_set_object (value,
		form_child->constraints[GTK_FORM_EDGE_BOTTOM].child ?
		G_OBJECT(form_child->constraints[GTK_FORM_EDGE_BOTTOM].child) : NULL);
    break;
  case CHILD_PROP_BOTTOM_OFFSET:
    g_value_set_uint(value, form_child->constraints[GTK_FORM_EDGE_BOTTOM].offset);
	break;
  case CHILD_PROP_RIGHT_ATTACHMENT:
    g_value_set_enum (value, form_child->constraints[GTK_FORM_EDGE_RIGHT].attachment);
	break;
  case CHILD_PROP_RIGHT_WIDGET:
    g_value_set_object (value,
		form_child->constraints[GTK_FORM_EDGE_RIGHT].child ?
		G_OBJECT(form_child->constraints[GTK_FORM_EDGE_RIGHT].child) : NULL);
    break;
  case CHILD_PROP_RIGHT_OFFSET:
    g_value_set_uint(value, form_child->constraints[GTK_FORM_EDGE_RIGHT].offset);
	break;
  default:
    GTK_CONTAINER_WARN_INVALID_CHILD_PROPERTY_ID (container, property_id, pspec);
    break;
  }
}

static void
gtk_form_init (GtkForm *form)
{
  GTK_WIDGET_SET_FLAGS (form, GTK_NO_WINDOW);
  gtk_widget_set_redraw_on_allocate (GTK_WIDGET (form), FALSE);
  
  form->children = NULL;
}

GtkWidget*
gtk_form_new ()
{
  GtkForm *form;

  form = g_object_new (GTK_TYPE_FORM, NULL);
  
  return GTK_WIDGET (form);
}

void
gtk_form_constrain (GtkForm	     *form,
		    GtkWidget	     *child,
		    GtkFormEdge       edge,
		    GtkFormAttachment attachment,
		    GtkWidget        *widget,
		    gint              offset)
{
  GtkFormChild *form_child;
  
  g_return_if_fail (form != NULL);
  g_return_if_fail (GTK_IS_FORM (form));
  g_return_if_fail (child != NULL);
  g_return_if_fail (GTK_IS_WIDGET (child));
  
  form_child = gtk_form_find_child (form, child);
  if (!form_child)
    return;

  form_child->constraints[edge].location = 0;
  form_child->constraints[edge].attachment = attachment;
  form_child->constraints[edge].offset = offset;
  form_child->constraints[edge].child = gtk_form_find_child (form, widget);
  
  if (GTK_WIDGET_VISIBLE (child->parent) /* && GTK_WIDGET_VISIBLE (child) */ )
    {
      if (GTK_WIDGET_MAPPED (child->parent))
	gtk_widget_map (child);

      gtk_widget_queue_resize (child);
    }
}

static void
gtk_form_finalize (GObject *object)
{
  GtkForm *form;
  
  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_FORM (object));
  
  form = GTK_FORM (object);
  
  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
gtk_form_size_request (GtkWidget      *widget,
		       GtkRequisition *requisition)
{
  GtkForm *form;
  GList *children;
  gint my_bounds [4] = { 0, 0, 0, 0 };
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_FORM (widget));
  g_return_if_fail (requisition != NULL);
  
  form = GTK_FORM (widget);
  
  for (children = form->children; children; children = children->next)
  {
      GtkRequisition aReq;
      GtkFormChild *child = children->data;

      gtk_widget_size_request (child->widget, &aReq);

      child->constraints [GTK_FORM_EDGE_RIGHT].location =
		child->constraints [GTK_FORM_EDGE_LEFT].location +
		aReq.width - 1;
      child->constraints [GTK_FORM_EDGE_BOTTOM].location =
		child->constraints [GTK_FORM_EDGE_TOP].location +
		aReq.height - 1;
  }

  gtk_form_layout (form, my_bounds, 1);

  requisition->width = my_bounds [GTK_FORM_EDGE_RIGHT] + 1;
  requisition->height = my_bounds [GTK_FORM_EDGE_BOTTOM] + 1;

  requisition->width += GTK_CONTAINER (form)->border_width * 2;
  requisition->height += GTK_CONTAINER (form)->border_width * 2;

  for (children = form->children; children; children = children->next)
  {
      GtkFormChild *child = children->data;

      child->constraints [GTK_FORM_EDGE_LEFT].fraction =
	    (float)child->constraints [GTK_FORM_EDGE_LEFT].location / 
		(float)(my_bounds [GTK_FORM_EDGE_RIGHT]);

      child->constraints [GTK_FORM_EDGE_RIGHT].fraction =
	    (float)child->constraints [GTK_FORM_EDGE_RIGHT].location / 
		(float)(my_bounds [GTK_FORM_EDGE_RIGHT]);

      child->constraints [GTK_FORM_EDGE_TOP].fraction =
	    (float)child->constraints [GTK_FORM_EDGE_TOP].location / 
		(float)(my_bounds [GTK_FORM_EDGE_BOTTOM]);

      child->constraints [GTK_FORM_EDGE_BOTTOM].fraction =
	    (float)child->constraints [GTK_FORM_EDGE_BOTTOM].location / 
		(float)(my_bounds [GTK_FORM_EDGE_BOTTOM]);
  }
}

static void
gtk_form_size_allocate (GtkWidget     *widget,
			GtkAllocation *allocation)
{
  GtkForm *form;
  GList *children;
  gint my_bounds [4];
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_FORM (widget));
  g_return_if_fail (allocation != NULL);
  
  widget->allocation = *allocation;
  form = GTK_FORM (widget);

  my_bounds [GTK_FORM_EDGE_TOP] = 0 ;
  my_bounds [GTK_FORM_EDGE_LEFT] = 0 ;
  my_bounds [GTK_FORM_EDGE_BOTTOM] = my_bounds [GTK_FORM_EDGE_TOP] +
  				     allocation->height - 1 - 2*GTK_CONTAINER (form)->border_width;
  my_bounds [GTK_FORM_EDGE_RIGHT] = my_bounds [GTK_FORM_EDGE_LEFT] +
  				    allocation->width - 1 - 2*GTK_CONTAINER (form)->border_width ;

  gtk_form_layout (form, my_bounds, 0);

  for (children = form->children; children; children = children->next)
  {
      GtkFormChild *child = children->data;
      GtkAllocation allocation;

      allocation.x = child->constraints [GTK_FORM_EDGE_LEFT].location +
				GTK_WIDGET (form)->allocation.x +
	  			GTK_CONTAINER (form)->border_width;
      allocation.y = child->constraints [GTK_FORM_EDGE_TOP].location +
				GTK_WIDGET (form)->allocation.y +
	  			GTK_CONTAINER (form)->border_width;
      allocation.width = child->constraints [GTK_FORM_EDGE_RIGHT].location -
      			 child->constraints [GTK_FORM_EDGE_LEFT].location + 1;
      allocation.height = child->constraints [GTK_FORM_EDGE_BOTTOM].location -
      			  child->constraints [GTK_FORM_EDGE_TOP].location + 1;

      gtk_widget_size_allocate (child->widget, &allocation);
  }
}

static void
gtk_form_add (GtkContainer *container,
	      GtkWidget    *widget)
{
  GtkForm *form;
  GtkFormChild *form_child;
  gint i;

  g_return_if_fail (container != NULL);
  g_return_if_fail (GTK_IS_FORM (container));
  g_return_if_fail (widget != NULL);
  g_return_if_fail (widget->parent == NULL);
  
  form = GTK_FORM (container);

  form_child = g_new (GtkFormChild, 1);
  form_child->widget = widget;

  for (i = 0; i < 4; i ++)
  {
    form_child->constraints[i].location = 0;
    form_child->constraints[i].attachment = GTK_FORM_ATTACH_NONE;
    form_child->constraints[i].offset = 0;
    form_child->constraints[i].factor = i < 2 ? 1 : -1;
    form_child->constraints[i].state = STATE_RESET;
    form_child->constraints[i].child = NULL;
    form_child->constraints[i].lower_container_relative = FALSE;
    form_child->constraints[i].fraction = 1.0;
  }
  
  form->children = g_list_prepend (form->children, form_child);
  
  gtk_widget_set_parent (widget, GTK_WIDGET (form));
}

static void
gtk_form_remove (GtkContainer *container,
		  GtkWidget    *widget)
{
  GtkForm *form;
  GtkFormChild *child;
  GList *children;
  
  g_return_if_fail (container != NULL);
  g_return_if_fail (GTK_IS_FORM (container));
  g_return_if_fail (widget != NULL);
  
  form = GTK_FORM (container);
  children = form->children;
  
  while (children)
    {
      child = children->data;
      children = children->next;
      
      if (child->widget == widget)
	{
	  gtk_widget_unparent (widget);
	  
	  form->children = g_list_remove (form->children, child);
	  g_free (child);
	  
	  if (GTK_WIDGET_VISIBLE (container))
	    gtk_widget_queue_resize (GTK_WIDGET (container));
	  break;
	}
    }
}

static void
gtk_form_forall (GtkContainer *container,
		  gboolean	include_internals,
		  GtkCallback	callback,
		  gpointer	callback_data)
{
  GtkForm *form;
  GtkFormChild *child;
  GList *children;
  
  g_return_if_fail (container != NULL);
  g_return_if_fail (GTK_IS_FORM (container));
  g_return_if_fail (callback != NULL);
  
  form = GTK_FORM (container);
  children = form->children;
  
  while (children)
    {
      child = children->data;
      children = children->next;
      
      (* callback) (child->widget, callback_data);
    }
}

static void
gtk_form_layout (GtkForm *form, gint *my_bounds, gboolean recompute_our_size)
{
    int count;
    for (count = 0; count < 10000; )
    {
        GList *list;
	gboolean it_worked = 1;

	// 
	// Reset the state of all edges.
	//

	for (list = form->children; list; list = list->next)
	{
	    int edge;
	    GtkFormChild *form_child = list->data;

	    for (edge = 0; edge < 4; edge ++)
	        form_child->constraints [edge].state = STATE_RESET;
	}

	for (list = form->children; list; list = list->next)
	{
	    GtkFormChild *form_child = list->data;

	    jmp_buf env;

	    if (setjmp (env) == 0)
		gtk_form_layout_child (env, form_child, my_bounds,
				       recompute_our_size);
	    else
	    {
		it_worked = 0;
		count ++;
		break;
	    }
	}

	if (it_worked)
	    break;
    }
}

static void
gtk_form_layout_child (jmp_buf env, GtkFormChild *fc, gint *my_bounds,
		       gboolean recompute_our_size)
{
    gint edge;
    for (edge = 0; edge < 4; edge ++)
	gtk_form_layout_edge (env, fc, edge, my_bounds, recompute_our_size);
}

static gint
gtk_form_layout_edge (jmp_buf env, GtkFormChild *fc, gint edge,
		      gint *my_bounds, gboolean recompute_our_size)
{
    GtkFormConstraint *ec = &fc->constraints [edge];

    if (ec->state != STATE_DONE)
    {
	if (ec->state == STATE_VISITED)
	    printf ("FormLayout.layout: Circular dependency!\n");
	else
	{
	    gint location = 0;

	    ec->state = STATE_VISITED;

	    /*
	     * At this point, we can do the work.
	     */

	    switch (ec->attachment)
	    {
		case GTK_FORM_ATTACH_SELF:
			if ( !recompute_our_size ) {
				location = (int)(my_bounds [edge] * ec->fraction);
				break ;
			}
			/* if computing size fall through to NONE */

		case GTK_FORM_ATTACH_NONE:
			location = ec->location;
			ec->offset = 0;

			/*
			 * Edges that don't have attachmenta are now
			 * treated as if they were attached relative to the
			 * opposite edge.  This lets us propagate the value
			 * for lower_container_relative.  Since it is possible
			 * that the opposite edge depends on this edge, we
			 * need to catch the circular dependency first.
			 */

			if ( fc->constraints[edge^2].state != STATE_VISITED ) {
				gtk_form_layout_edge(env, fc, edge^2, my_bounds,
									recompute_our_size);
				location = ec->location ;
			}
			ec->lower_container_relative = 
				fc->constraints[edge^2].lower_container_relative;
		    break;

		case GTK_FORM_ATTACH_FORM:
		    location = my_bounds [edge];
			ec->lower_container_relative =
					edge == GTK_FORM_EDGE_BOTTOM ||
					edge == GTK_FORM_EDGE_RIGHT ;
		    break;

		case GTK_FORM_ATTACH_WIDGET:
		    location = ec->factor +	/* This IS correct */
			       gtk_form_layout_edge (env, ec->child,
						     edge ^ 2, my_bounds,
						     recompute_our_size);
			ec->lower_container_relative =
				fc->constraints[edge^2].lower_container_relative;
		    break;

		case GTK_FORM_ATTACH_OPPOSITE_WIDGET:
		    location = gtk_form_layout_edge (env, ec->child,
						     edge, my_bounds,
						     recompute_our_size);
			ec->lower_container_relative =
				fc->constraints[edge].lower_container_relative;
		    break;

		case GTK_FORM_ATTACH_CENTER:
		{
		    gint center;
		    gint size;

		    if (ec->child == NULL)	/* Center on form */
		    {
			center = (my_bounds [edge ^ 2] -
				  my_bounds [edge]) / 2;
			if (center < 0) center = -center;
		    }
		    else
		    {
		        gint edge1;
		        gint edge2;
		        gint half;

			gtk_form_layout_edge (env, ec->child,
					      edge ^ 2, my_bounds,
					      recompute_our_size);
 			gtk_form_layout_edge (env, ec->child,
					      edge, my_bounds,
					      recompute_our_size);

			/* It might be tempting to use the return of
			 * gtk_form_layout_edge rather than the next 2 lines
			 * of code.  Bad idea because the second call to
			 * gtk_form_layout_edge may move the first edge.
			 */

			edge1 = ec->child->constraints [edge^2].location;
			edge2 = ec->child->constraints [edge].location;
			half = (edge1 - edge2) / 2;
			center = ec->child->constraints [edge].location +
				 half;
		    }

		    /* We depend on the opposite edge so lets lay him out
		     * first.
			 */

		    gtk_form_layout_edge (env, fc, edge ^ 2, my_bounds,
					    recompute_our_size);

		    size = fc->constraints [edge ^ 2].location -
			       fc->constraints [edge].location + 1;

		    location = center - size / 2;

		    break;
		}

		default:
		    printf ("FormLayout: Unknown attachment type!\n");
	    }

	    location += ec->offset * ec->factor;

	    gtk_form_move_edge (env, fc, edge, location, my_bounds,
				recompute_our_size && ec->lower_container_relative);

	    ec->state = STATE_DONE;
	}
    }

    return fc->constraints [edge].location;
}

static void
gtk_form_move_edge (jmp_buf env, GtkFormChild *fc, int edge, int where,
		    gint *my_bounds, gboolean recompute)
{
    int diff = where - fc->constraints [edge].location;

    int opposite_edge = edge ^ 2;

    if (gtk_form_edge_should_move_too (fc, opposite_edge))
	fc->constraints [opposite_edge].location += diff;
    else
    {
	/* Special Case:  If we are in "recompute" mode and we shrink
	 * because of this constraint, then we need to expand the form
	 * to accomodate us instead of shrinking the child.
	 */

	if (recompute)
	{
	    int delta = diff * fc->constraints [opposite_edge].factor;

	    if (delta < 0)
	    {
		if (edge == GTK_FORM_EDGE_TOP || edge == GTK_FORM_EDGE_BOTTOM)
		    my_bounds [GTK_FORM_EDGE_BOTTOM] += -delta;
		else
		    my_bounds [GTK_FORM_EDGE_RIGHT] += -delta;
		
		longjmp (env, 1);
	    }
	}
    }

    fc->constraints [edge].location += diff;
}

static gboolean
gtk_form_edge_should_move_too (GtkFormChild *fc, gint edge)
{
    /* Should edge move if the opposite edge moves.  e.g. If the LEFT
     * edge moves 10 pixels, should the RIGHT edge move too?
	 *
     * Simply stated, an edge can move if there are no constraints for
     * that edge or the constraints for that edge are relative to itself
     * (specifies a width) or the edge has not been layed out yet.
	 */

    int attachment = fc->constraints [edge].attachment;

    return (attachment == GTK_FORM_ATTACH_NONE) ||
	   (fc->constraints [edge].state != STATE_DONE) ||
	   (attachment == GTK_FORM_ATTACH_WIDGET &&
	    fc->constraints [edge].child == fc);
}
